/*  File   : core.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Core of <language>-to-<formatter> program.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 core.c	1.32";
#endif/*lint*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "many2h.h"
#include "language.h"
#include "formatter.h"


/*  For reporting errors, we want to keep track of which file we are
    reading and what line we are on.  We also want to keep track of
    errors.
*/
static int         final_status = EXIT_SUCCESS;
static char const *file_name;
static long        line_number;
static int	   numbering = 0;

void where_am_i(int column) {
    if (file_name != 0) {
	fprintf(stderr, "%s:", file_name);
    }
    if (line_number != 0) {
	fprintf(stderr, "%ld:", line_number);
	if (column >= 0) fprintf(stderr, "%d:", column+1);
    }
    final_status = EXIT_FAILURE;
}


/*  I want a FOO-to-HTML converter that can be used with
    - Ada
    - AWK
    - C
    - C++
    - Clean   (not yet done)
    - Eiffel
    - Erlang
    - Haskell (not yet done)
    - HTML
    - Java
    - Lisp    (not yet done)
    - Mercury
    - Pascal
    - Prolog
    - Scheme  (not yet done)
    - SGML
    - Standard ML (not quite right)
    - SQL
    - XBase   (not yet done)

    There are two reasons for the (not yet done) entries.
    * XBase is too messy
    * the others have nested comments.
    Once any one of the nested comment languages is done,
    the others will follow fast.

    This thing is to preserve the line breaks.
    It is to add markup and convert HTML-special characters.
    It may also adjust spacing.

    A line is a sequence of tokens.
    A token is
	a) an Attribute.
	   An attribute is an identifier following the Tick symbol (Ada)
	   or preceding (DSSSL id:) or following (Lisp :id) a special mark.
	   There may be a fixed set of attributes as in Ada, or they may
	   be user-defined as in Lisp, but they are not keywords, not being
	   reserved in all contexts, and they are not identifiers or field
	   names, not having values.  If an otherwise reserved word occurs
	   in an attribute context, it is treated as an attribute.

	k) a Keyword.
	   A keyword is an identifier which has special syntactic properties;
	   it is defined by the language and is typically reserved in all
	   contexts.  There are two kinds of keywords:  M4 keywords which
	   are such only if they are immediately followed by a left
	   parenthesis, and language keywords, which should be separated
	   from any following bracket.

	   It might be a good idea in Pascal for
	   <k:end> <u> <v:kw> ... where the comment part kw is a keyword,
	   to be rendered as <k:end> <u> <k:kw> ....

	f) a Field name.
	   A field name is an identifier that stands for a field in a
	   record; it is recognised by following a dot [.x ->x] or
	   preceding an arrow  [(x=> ,x=>].  These simple rules do NOT
	   handle the defining occurrences of field names.  If a reserved
	   word occurs in a field name context, it is taken as a reserved
	   word; so far the only instance is .all in Ada.

	   We can recognise defining occurrences of field names in Ada
	   and Pascal by looking for identifiers between 'record' and
	   'end', preceding a colon.  Roughly:
		record_ct := 0, record_fd := false
		'record' => record_ct +:= 1, record_fd := true
		'case' & record_ct > 0 => record_ct +:= 1
		'end'  & record_ct > 0 => record_ct -:= 1
		';'    & record_ct > 0 => record_fd := true
		':'    & record_ct > 0 => record_fd := false
		id     & record_fd     => is_field.
	   This has its imperfections.  It won't work for C.

	i) an Identifier.
	   An identifier is a word that is user defined.  It is recognised
	   by not being any other kind of word.

	c) a Character name.
	   This token class is set up for Lisp, with \Control-Space and
	   the like.  A Character-name has the form
	    \ {Word-}*{Word | Char}
	   where a Word is a sequence of two or more letters.

	n) a Number
	   A number begins with a digit.

	l) a Left bracket, typically ( [ {
	
	r) a Right bracket, typically ) ] }

	m) a coMMa, typically , ; : as all should be followed by space
	   In Ada and Pascal, a colon may be preceded by any number of
	   spaces for alignment.

	d) a Dot, typically . but including -> in C

	t) a Tick, introducing an attribute

	o) a string Opener
	p) a string Part
	q) a string Quitter


	u) a comment opener (comments help Understanding)
	v) a comment part
	w) a comment closer

	s) any other Special character

	e) the End of the line.

    The letters remaining for additional kinds of token are bhjxyz.

    This is not the definition used by the languages themselves.
    For example,
	if x /= ' ' then -- Ada example
    corresponds to the tokens
	k<if> i<x> s</> s<=> o<'> p< > q<'> k<then> u<-- > v<Ada example> w<>
    Each token has
	- a type
	- a starting column
	- a length
	- a text
    where the text is represented by a pointer into a buffer.

    Some tokens may be treated as keywords or identifiers depending on
    style, notably Boolean, True, and False in Pascal and Ada.  In the
    relevant language standards, they are defined as ordinary identifiers.
    I had considered having a command line switch to allow "boolean" ids
    to be treated as keywords, but a much better scheme is simply to let
    the user specify a list of extra keywords.  That will also handle
    extensions.
*/

struct Token_Looks {
    enum Text_Style style;	/* font face */
    int             wanted;	/* is it wanted at all */
                                /* case conversion */
};
static struct Token_Looks looks[] = {
    { Plain,  0 },		/* e. End_Of_Line */
    { Bold,   1 },		/* k. Keyword */
    { Plain,  1 },		/* a. Attribute */
    { Plain,  1 },		/* f. Field_Name */
    { Plain,  1 },		/* i. Identifier */
    { Typed,  1 },		/* c. Character_Name */
    { Plain,  1 },		/* n. Number */
    { Plain,  1 },		/* l. Left_Bracket */
    { Plain,  1 },		/* r. Right_Bracket */
    { Plain,  1 },		/* m. Comma */
    { Plain,  1 },		/* d. Dot */
    { Plain,  1 },		/* t. Tick */
    { Plain,  1 },		/* o. String_Opener */
    { Typed,  1 },		/* p. String_Element */
    { Plain,  1 },		/* q. String_Closer */
    { Plain,  1 },		/* u. Comment_Opener */
    { Italic, 1 },		/* v. Comment_Element */
    { Plain,  1 },		/* w. Comment_Closer */
    { Plain,  1 }		/* s. Special */
};


static void token_map(
    char const *s,
    int         v,
    void      (*f)(enum Token_Type, int)
) {
    enum Token_Type l, u;
    /*  Alternatives:
	  '.' = 'f' (field name)
	  '&' = 'i' (HTML/XML/SGML identifier)
	  '#' = 'x' (comment parts)
	Ranges:
	  'j' = Attribute, Field_Name, Identifier, Character_Name
	  'x' = Comment_Opener, Comment_Element, Comment_Closer
	  '@' = String_Opener, String_Element, String_Closer
	  '*' = everything except End_Of_Line
	Unused letters:
	  b, e, g, h, y, z
    */

    for (;;) {
	switch (*s++) {
	case '\0':                                   return;
	case 'a': case 'A':  l = u = Attribute;      break;
	case 'c': case 'C':  l = u = Character_Name; break;
	case 'd': case 'D':  l = u = Dot;            break;
	case '.':
	case 'f': case 'F':  l = u = Field_Name;     break;
	case '&':
	case 'i': case 'I':  l = u = Identifier;     break;
	case 'j': case 'J':  l =     Attribute;
	                         u = Character_Name; break;
	case 'k': case 'K':  l = u = Keyword;        break;
	case 'l': case 'L':  l = u = Left_Bracket;   break;
	case 'm': case 'M':  l = u = Comma;          break;
	case 'n': case 'N':  l = u = Number;         break;
	case 'o': case 'O':  l = u = String_Opener;  break;
	case 'p': case 'P':  l = u = String_Element; break;
	case 'q': case 'Q':  l = u = String_Closer;  break;
	case 'r': case 'R':  l = u = Right_Bracket;  break;
	case 's': case 'S':  l = u = Special;        break;
	case 't': case 'T':  l = u = Tick;           break;
	case 'u': case 'U':  l = u = Comment_Opener; break;
	case 'v': case 'V':  l = u = Comment_Element;break;
	case 'w': case 'W':  l = u = Comment_Closer; break;
	case 'x': case 'X':
	case '#':            l =     Comment_Opener;
	                         u = Comment_Closer; break;
	case '@':            l =     String_Opener;
	                         u = String_Closer;  break;
	case '*':	     l =     Keyword;
	                         u = Special;        break;
	default:             l =     Special;
	                         u = Keyword;      break;
	}
	for (; l <= u; l++) f(l, v);
    }
}

static void a_set_wanted(enum Token_Type t, int v) {
    looks[t].wanted = v;
}

static void set_wanted(char const *s, int to) {
    token_map(s, to, a_set_wanted);
}

static void a_set_style(enum Token_Type t, int v) {
    looks[t].style = (enum Text_Style)v;
}

static void set_style(char const *s, enum Text_Style style) {
    token_map(s, (int)style, a_set_style);
}

static void a_add_style(enum Token_Type t, int v) {
    looks[t].style |= (enum Text_Style)v;
}

static void add_style(char const *s, enum Text_Style style) {
    token_map(s, (int)style, a_add_style);
}


#define	CS_verbatim	0	/* Leave letter unchanged */
#define CS_lower	1	/* Change letter to lower case */
#define	CS_upper	2	/* Change letter to upper case */
#define CS_title	3	/* Change letter to title case */

struct Case_Shift {
    unsigned int c11 : 2;	/* Abbb_cddd */
    unsigned int c1m : 2;	/* aBBB_cddd */
    unsigned int cn1 : 2;	/* abbb_Cddd */
    unsigned int cnm : 2;	/* abbb_cDDD */
    unsigned char sep;		/* what boundary marker to use */
};

static struct Case_Shift case_shift[] = {
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
    { CS_verbatim, CS_verbatim, CS_verbatim, CS_verbatim, '\0' },
};

static void a_set_case(enum Token_Type t, int v) {
    case_shift[t].c11 = v & 3;
    case_shift[t].c1m = (v >> 2) & 3;
    case_shift[t].cn1 = (v >> 4) & 3;
    case_shift[t].cnm = (v >> 6) & 3;
    case_shift[t].sep = (v >> 8) & 255;
}

static void set_case_shift(char const *s, int to) {
    token_map(s, to, a_set_case);
}

/*  Between lines, the tokeniser can be in the following states:
    - the last token was one that can end a term
    - the last token was a tick
    - in the middle of a string (looking for a specified terminator)
    - in the middle of a comment (looking for a specified terminator)
    - anything else.
*/
enum Parser_State old_state = Other_Token;
enum Parser_State state     = Other_Token;
int               lq_count  = 0;

unsigned char const *
    terminator = (unsigned char const *)"";
size_t
    termlength = 0;


/*  We need to look for a terminator in the current line.
    As a special case, "" is found at the end of the line.
    We don't use strstr(), because we want to ignore the escape.
    The escape character should be ignored in strings, but
    not in comments.  For the moment, we don't handle nesting
    comments or nesting strings.

    To handle languages like Fortran, Pascal, and Smalltalk
    where an apostrophe inside a string is rendered as '',
    we take escape==quote to mean quotes should be doubled.
*/
int terminator_column(int column, int escape) {
    if (escape == '\0') {                /* There is no escape character */
	do {
	    if (line_buffer[column] == terminator[0]
	     && memcmp(&line_buffer[column+1], &terminator[1], termlength-1) == 0
	    ) {
		return column;
	    }
	} while (line_buffer[column++] != '\0');
    } else
    if (escape == terminator[0]) {	/* Quotes are doubled */
	while (line_buffer[column] != '\0') {
	    if (line_buffer[column] == escape) {
		if (line_buffer[column+1] != escape) return column;
		column += 2;
	    } else {
	        column += 1;
	    }
	}
    } else {				/* Backslash escapes */
	do {
	    if (line_buffer[column] == terminator[0]
	     && memcmp(&line_buffer[column+1], &terminator[1], termlength-1) == 0
	    ) {
		return column;
	    }
	    if (line_buffer[column] == escape) column++;
	} while (line_buffer[column++] != '\0');
    }
    return -1;
}


static struct Formatter const *fmt;

static void
debug_token_row(void) {
    int t;
    static char const k[] = "ekaficnlrmdtopquvws";
    static char const s[] = "epscox";

    printf("%c", s[(int)state]);
    for (t = 0; token_row[t].type != End_Of_Line; t++) {
	struct Token *n = &token_row[t];
	printf(" %c[%.*s]",
	       k[(int)n->type], (int)n->length, (char const *)n->text);
    }
    printf("\n");
}
	

/*  SLOC counting  */

static long orig_total = 0;
static long sloc_total = 0;
static int  want_sloc  = 0;
static int  want_total = 0;

static int line_is_interesting(void) {
    int t;

    for (t = 0; token_row[t].type != End_Of_Line; t++) {
        switch (token_row[t].type) {
            case Attribute:
            case Field_Name:
            case Identifier:
            case Character_Name:
            case Number:
            case String_Opener:
                return 1;
	    default:
	        break;
	}
    }
    return 0;
}

/*  IMPLEMENT SOON:
    This is the place to do line breaking, for the sake of crud
    like the Jikes compiler with its excessively long lines.
    To do line breaking, we have to know how wide the output
    should be, so we'll also need a -w# option.
*/
static void
write_token_row(void) {
    int column, next_column;
    int t;

    column = 0;
    for (t = 0; token_row[t].type != End_Of_Line; t++) {
	struct Token *n = &token_row[t];

	next_column = n->start + n->length;
	if (n->length != 0) {
	    if (looks[n->type].wanted) {
		fmt->tab_from_to(column, n->start);
		write_chunk(n->text, n->length, fmt, looks[n->type].style);
		fmt->end_token(n->type);
	    } else {
		fmt->tab_from_to(column, next_column);
	    }
	    column = next_column;
	}
    }
    fmt->finish_line();
}


/*  We want to adjust spacing in the following cases:

 -  an Identifier, Attribute, or Tick followed by ( or [
    should not have any space in between.

 -  a Keyword followed by ( [ or { should have at least one space in between.

 -  a { (C,C++,Java,AWK,Hawk) should have at least one space between it and
    the previous Identifier or Keyword.

 -  a Comma should have at least one space after it, and generally,
    no space before it.

 -  there should be exactly as much space before a Right_Bracket
    as there is after the corresponding Left_Bracket.

    This is the first draft of this code, and it just keeps track of
    the adjustments to be made.
*/

#define BRACKET_LIMIT 10000
static unsigned bracket_space[BRACKET_LIMIT];
static unsigned bracket_level = 0;

static void
begin_spacing(void) {
    bracket_level = 0;
}

static void
end_spacing(void) {
    if (bracket_level != 0) {
	fprintf(stderr, "%u bracket levels still open\n", bracket_level);
    }
}

static void
do_case_shift(void) {
    int t;		/* token_row[] index */

    for (t = 0; token_row[t].type != End_Of_Line; t++) {
	register struct Case_Shift *cs = &case_shift[token_row[t].type];
	unsigned char *p = (unsigned char *)token_row[t].text;
	unsigned char *e = token_row[t].length + p;
	int seen_syll = 0;	/* seen any syllable of word yet? */
	int seen_char = 0;	/* seen any letter of syllable yet? */
	unsigned char c;
	unsigned int m;

	for (; p < e; p++) {
	    c = *p;
	    if (isalpha(c)) {
		if (seen_syll) {
		    if (seen_char) {
			m = cs->cnm;
		    } else {
			seen_char = 1;
			m = cs->cn1;
		    }
		} else {
		    if (seen_char) {
			m = cs->c1m;
		    } else {
			seen_char = 1;
			m = cs->c11;
		    }
		}
		switch (m) {
		    case CS_lower:	*p = tolower(c); break;
		    case CS_upper:	*p = toupper(c); break;
		    case CS_title:	*p = toupper(c); break;
		    default:				 break;
		}
	    } else
	    if (c <= ' ') {
		seen_syll = seen_char = 0;
	    } else {
		if (seen_char) seen_syll = 1, seen_char = 0;
	    }
	}
    }
}

static void
adjust_spacing(void) {
    int t;
    int prev_end, this_end;
    int delta;

    prev_end = 0;
    for (t = 0; token_row[t].type != End_Of_Line; t++) {
	this_end = token_row[t].start + token_row[t].length;
	switch (token_row[t].type) {
	    case Left_Bracket:
		bracket_space[bracket_level++] =
		    token_row[t+1].start - this_end;
		break;
	    case Right_Bracket:
		if (bracket_level == 0) {
		    where_am_i(token_row[t].start);
		    fprintf(stderr, "Bracketing error\n");
		} else {
		    bracket_level--;
		    delta = token_row[t].start - prev_end
			  - bracket_space[bracket_level];
		    if (t > 0 && delta > 0) {
			put_string("])");
		    } else
		    if (t > 0 && delta < 0) {
			put_string("[)");
		    }
		}
		break;
	    case Identifier:
	    case Field_Name:
	    case Attribute:
	    case String_Closer:
	    case Dot:
	    case Tick:
		if (token_row[t+1].type == Left_Bracket) {
		    delta = token_row[t+1].start - this_end;
		    assert(delta >= 0);
		    if (token_row[t+1].text[0] == '{') {
		        /* separating space is required before BEGIN */
			if (delta == 0) put_string("|(");
		    } else {
			/* separating space is forbidden before ([ */
			if (delta >  0) put_string("_(");
		    }
		}
		break;
	    case Keyword:
		if (token_row[t+1].type == Left_Bracket) {
		    delta = token_row[t+1].start - this_end;
		    assert(delta >= 0);
		    if (is_keyword(&token_row[t]) == 2) {
			/* M4 keywords forbid separating space */
			if (delta >  0) put_string("_(");
		    } else {
			/* plain keywords require separating space */
			if (delta == 0) put_string("|(");
		    }
		}
		break;
	    case Comma:
		if (t > 0 && token_row[t].text[0] != ':') {
		    /* <id> <spaces> : is allowed in Ada and Pascal */
		    /* declarations; we don't want to allow it in   */
		    /* statement labels, but how do we discriminate?*/
		    delta = token_row[t].start - prev_end;
		    assert(delta >= 0);
		    if (delta > 0) put_string("_,");
		}
		if (token_row[t+1].type != End_Of_Line) {
		    delta = token_row[t+1].start - prev_end;
		    assert(delta >= 0);
		    if (delta == 0) put_string(",|");
		}
		break;
	    default:
		break;
	}
	prev_end = this_end;
    }
}
		    

static struct Language const *cur_lang, *set_lang;

static int debug = 0;
static int some_done = 0;

static void
process(char const *title) {
    FILE *stream;

    if (!some_done) {
	if (strcmp(title, "-") == 0) {
	    fmt->begin_output("", "", cur_lang->name, " source code");
	} else {
	    fmt->begin_output("", "", title, " and more");
	}
	some_done = 1;
    }

    file_name = title;
    if (strcmp(title, "-") == 0) {
	stream = stdin;
	fmt->begin_section("");
    } else
    if ((stream = fopen(title, "rb")) == 0) {
	fprintf(stderr, "%s: cannot open.\n", title);
	final_status = EXIT_FAILURE;
	return;
    } else {
	fmt->begin_section(title);
    }
    old_state = Other_Token;
    state     = Other_Token;
    lq_count  = 0;
    if (want_sloc) {
	long sloc_lines = 0;
	long orig_lines = 0;

	while (read_line(stream)) {
	    orig_lines++;
	    cur_lang->tokenise();
	    if (line_is_interesting()) {
		sloc_lines++;
	        if (numbering) printf("%5ld.  %s\n", orig_lines, line_buffer);
	    }
	}
	sloc_total += sloc_lines;
	orig_total += orig_lines;
	printf("%7ld %7ld %s\n", sloc_lines, orig_lines, title);
    } else {
	begin_spacing();
	for (line_number = 1; read_line(stream); line_number++) {
	    cur_lang->tokenise();
	    /*adjust_spacing();*/
	    do_case_shift();
	    if (numbering) printf("%54ld.  ", line_number);
	    if (debug) debug_token_row(); else
	    write_token_row();
	}
	end_spacing();
    }
    if (stream != stdin) {
	fclose(stream);
    }
    fmt->end_section();
}


/*  This is not the official interface yet.
    There is plenty of work still to do before that.
    For now, it is

	many2h [-m] [-n] [-l$L] [-f$F] [-k$K] [-i$T|-e$T...] [file]...

    -m switches m4 keyword recognition ON.
    -n switches line numbering ON.
    -l$L specifies source language $L, currently
	ada, c, c++, eiffel, erlang, html, java, mercury, pascal,
        prolog, raw, sgml, xml
	If -l is not provided, the language is guessed from the
	file extension.  If there isn't any file either, 'raw'
	is guessed for stdin.
    -f$ specifies formatter $F, currently
	html, coloured html, latex, plain, ansi, coloured ansi.
	If not specified, plain is assumed.
        -fs specifies sloc counting.  The output is
	<SLOC> <ORIG> <FILE> for each file name, where
	<SLOC> is the number of interesting lines, <ORIG> is the
	number of lines originally, and <FILE> is the file name.
        -fst asks for a summary <SLOC> <ORIG> total line as well.
	A line is interesting if and only it contains at least
	one Attribute, Field_Name, Identifier, Character_Name,
	Number, or String_Opener token.  I am not sure about '#' lines.
    -k$K specifies additional keywords $K=kw,kw,...
	Changing language clears out keywords, so if you use this,
	put it _after_ -l.  As a special hack, it will look ahead
	to the following argument (which should be a file name) if
	there is no -l.
    -i$T says to include only the indicated tokens in the output.
	For example, to include only comment bodies,   do -iv
    -e$T says to exclude the indicated tokens from the output.
	For example, to exclude all parts of comments, do -ex
    -ad Sets default style (plain, keyword bf, strings tt, comments it)
    -ae Sets Eiffel style (comments plain, identifiers it)
    -a$S$T Sets the style $S for the tokens $T where $S is
	p(lain), i(talic), b(old), u(nderlined), or t(ypewriter).
    -x$N tab eXpansion uses $N column tabstop placement.
        If $N is 0, or the option is not used, N is taken as 8.
*/
int main(int argc, char **argv) {
    int i;
    int M4_used = 0;
    char const *p;
#   define set_p_to_argument_text { 				\
	p = argv[i]+2;						\
	if (*p == '/' || *p == ':' || *p == '=') p++; else	\
	if (*p == '\0' && i+1 < argc) p = argv[++i];		\
    }

    fmt = &plain;
    set_lang = 0;
    init_ctype(M4_used, cur_lang = &raw);

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-') {
            switch (argv[i][1]) {
            case 'd': case 'D':
                debug = 1;
                break;
            case 'm': case 'M':
                M4_used = 1;
                init_ctype(M4_used, cur_lang = set_lang);
                break;
	    case 'n': case 'N':
	        numbering = 1;
	        break;
	    case 'l': case 'L':
		set_p_to_argument_text
	        set_lang = opt_lang(p);
	        init_ctype(M4_used, cur_lang = set_lang);
	        break;
	    case 'k': case 'K':
		set_p_to_argument_text
		if (set_lang == 0 && i+1 < argc && argv[i+1][0] == '-') {
		    set_lang = ext_lang(argv[i+1]);
		    init_ctype(M4_used, cur_lang = set_lang);
		}
		add_user_keywords(p);
		break;
	    case 'f': case 'F':
		set_p_to_argument_text
		if ((p[0]|32) == 's') {
		    want_sloc = 1;
		    if ((p[1]|32) == 't') want_total = 1;
		    p = "p";
		}
	        fmt = opt_fmt(p);
		break;
	    case 'i': case 'I':
		set_p_to_argument_text
		set_wanted("*", 0);
		set_wanted(p, 1);
		break;
	    case 'e': case 'E':
		set_p_to_argument_text
		set_wanted(p, 0);
		break;
	    case 'x': case 'X':
	        set_p_to_argument_text
	        tab_width = atoi(p);
	        if (tab_width == 0) tab_width = 8;
	        if (tab_width < 0 || tab_width > 20) {
	            fprintf(stderr, "m2h: -xN requires 0<=N<=20\n");
	            exit(EXIT_FAILURE);
		}
		break;
	    case 'c': case 'C':
		set_p_to_argument_text
		{
		    unsigned int c11, c1m, cn1, cnm, sep;

		    switch (*p) {
			case '0':	case '-':
			case '_':	case ' ':
			    sep = *p++;
			    break;
			default:
			    sep = '\0';
			    break;
		    }
		    switch (*p++) {
		        case 'v': case 'V':
		            c11 = c1m = cn1 = cnm = CS_verbatim;
		            break;
			case 'l': case 'L':
		            c11 = c1m = cn1 = cnm = CS_lower;
		            break;
			case 'u': case 'U':
			    c11 = c1m = cn1 = cnm = CS_upper;
			    break;
			case 't': case 'T':
			    c11 = cn1 = CS_title;
			    c1m = cnm = CS_lower;
			    break;
			case 'b': case 'B':
			    cn1 = CS_title;
			    c11 = c1m = cnm = CS_lower;
			    break;
			case 'm': case 'M':
			    c11 = CS_verbatim;
			    cn1 = CS_title;
			    c1m = cnm = CS_lower;
			    break;
			case 'x': case 'X':
			    c11 = cn1 = CS_verbatim;
			    c1m = cnm = CS_lower; 
			    break;
			default:
			    fprintf(stderr, "m2h: -c %s unknown suboption\n",
				    p-1);
			    fprintf(stderr, "    -cv	verbatim\n");
			    fprintf(stderr, "    -cl	lower case\n");
			    fprintf(stderr, "    -cu	UPPER CASE\n");
			    fprintf(stderr, "    -ct	Title Case\n");
			    fprintf(stderr, "    -cb	baStudlyCaps\n");
			    fprintf(stderr, "    -cm	Mixed case\n");
			    fprintf(stderr, "    -cx	(obsolete)\n");
			    exit(EXIT_FAILURE);
		    }
		    if (sep != '\0') {
			fprintf(stderr, "m2h: -c%c%s  %s\n",
			    sep, p, "boundary change not implemented yet");
			exit(EXIT_FAILURE);
		    }
		    if (*p == '\0') p = "aif";
		    set_case_shift(p, c11 | (c1m << 2) | 
			(cn1 << 4) | (cnm << 6) | sep << 8);
		}
		break;
	    case 'a': case 'A':
		set_p_to_argument_text
		switch (*p) {
		    case 'c': case 'C':
			looks[Comment_Opener].style =			
			looks[Comment_Closer].style =
			    looks[Comment_Element].style;
			break;
		    case 'd': case 'D':
		        set_style("*",  Plain);
		        set_style("k",  Bold);
		        set_style("v",  Italic);
		        set_style("cp", Typed);
		        break;
		    case 'e': case 'E':
		        set_style("*",  Plain);
		        set_style("k",  Bold);
		        set_style("j",  Italic);
		        set_style("cp", Typed);
		        break;
		    case 'p': case 'P':
		        set_style(p+1, Plain);
		        break;
		    case 'b': case 'B':
		        set_style(p+1, Bold);
		        break;
		    case 'i': case 'I':
		        set_style(p+1, Italic);
		        break;
		    case 't': case 'T':
		        set_style(p+1, Typed);
		        break;
		    case 'u': case 'U':
		        set_style(p+1, Ulined);
		        break;
		    default:
		        fprintf(stderr, "m2h: unknown appearance: -a%s\n", p);
		        fprintf(stderr, "-ad  (Default)\n");
		        fprintf(stderr, "-ae  (Eiffel)\n");
		        fprintf(stderr, "-apX (Plain for tokens X)\n");
		        fprintf(stderr, "-abX (Bold for tokens X)\n");
		        fprintf(stderr, "-aiX (Italic for tokens X)\n");
		        fprintf(stderr, "-auX (Underline for tokens X)\n");
		        fprintf(stderr, "-atX (Typewriter for tokens X)\n");
        		exit(EXIT_FAILURE);
		}
		break;
	    case 't': case 'T':
	        add_style("*", Typed);
	        break;
	    case '\0':
	        process("-");
	        some_done = 1;
	        break;
	    default:
		fprintf(stderr, "m2h: unknown option: %s\n", argv[i]);
		exit(EXIT_FAILURE);
	    }		
	} else {
	    if (set_lang == 0) {
		init_ctype(M4_used, cur_lang = ext_lang(argv[i]));
	    }
	    process(argv[i]);
	}
    }

    if (!some_done) process("-");
    if (want_total) printf("%7ld %7ld\n", sloc_total, orig_total);
    fmt->end_output();
    flush_output();
    return final_status;
}

/*  Still to be done:

    case shifting (yes, it WOULD apply to numbers)
	-c$B?$S$T
	    $B says what to do about word boundaries.
	    A word boundary occurs in a token whenever
	    + an upper case letter follows a lower case letter (b="")
	    + any letter follows a hyphen (b="-")
	    + any letter follows an underscore (b="_")
	    + any letter follows a blank (b=" ")
	    If $B is missing, don't do anything.
	    If $B is "0", replace the found boundary (b) by nothing.
	    If $B is "-", "_", or " ", replace by found boundary (b)
	    	by $B.
	    
	    $S says what to do about case shifting.
	    v	means take verbatim; change nothing
	    l	means convert to lower case
	    u	means convert to UPPER case
	    t	means convert to Title case 
	    b	means convert to baStudlyCaps (lc initial, uc after break)
	    m	means convert to mixed case retaining current initial

	    $T is the set of tokens that this conversion is to apply to.
	    If $T is empty, the default is "aif".

	The snag with this is that word boundary changes may force
	layout adjustment.  My current idea is to maintain a running
	delta, and whenever <comma><space>* or <space>+ is found, to
	make a local adjustment, trying to preserve as much alignment
	as possible.

    layout adjustment
    - to correct tabbing at the start of a line
    - to fiddle with parentheses and commas (so Comma should be a token
      too, and ,;: would all be in that class).
    - to align comments

    special-casing 'syllables' of identifiers.
*/

