/*  File   : language.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Shared tokenising code.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 language.c	1.24";
#endif/*lint*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "many2h.h"
#include "language.h"

unsigned char tocase[3][1 + (int)UCHAR_MAX];
char ctype[256];
char escaper = '\0';	/* Use \0 to mean none. */
char opchar  = '\0';    /* Use \0 to mean none. */

/*  All the languages agree that
    - 0 to 9 are DIGITs
    - the plain and accented lower case letters are LOWERs
    - the plain and accented upper case letters are UPPERs
    - other characters may appear in identifiers, typically "_"
    - some subset of ([{ are LBRACKs
    - some subset of }]) are RBRACKs
    - some subset of the remaining characters are SPECLs
*/

/*  An identifier may be
    <0 a comment keyword (NOTE in xbase, dnl in M4, comment in Algol 68),
    =0 just an identifier.
    =1 a plain keyword (everything else)
    =2 an M4 keyword that wants an immediately following (
*/
struct Keyword_Tail {
    struct Keyword_Tail *next;
    unsigned char const *tail;
    size_t               size;
    int                  kind;
};

struct Keyword_Set {
    enum CsSns           case_sensitive;
    struct Keyword_Tail *words[1 + UCHAR_MAX];
};

static struct Keyword_Set keyword_set;


void
add_kinded_keyword(char const *word, size_t size, int kind) {
    unsigned char const *uword = (unsigned char const *)word;
    struct Keyword_Tail *link;

    link = malloc(sizeof *link);
    if (link == 0) {
        fprintf(stderr, "Cannot build keyword table.\n");
        exit(EXIT_FAILURE);
    }
    link->next = keyword_set.words[uword[0]];
    link->size = size - 1;
    link->tail = uword + 1;
    link->kind = kind;
    keyword_set.words[uword[0]] = link;
}


static void
add_keyword(char const *word) {
    add_kinded_keyword(word, strlen(word), 1);
}


static void
add_m4_keyword(char const *word) {
    add_kinded_keyword(word, strlen(word), 2);
}


void add_user_keywords(char const *words) {
    unsigned char const *f = (unsigned char const *)words;
    unsigned char const *s;

    while (*f != '\0' && !is_a_letter(*f)) f++;
    while (*f != '\0') {
	s = f;
	while (is_a_letter(*f)) f++;
	add_kinded_keyword((char const *)s, f-s, 1);
	while (*f != '\0' && !is_a_letter(*f)) f++;
    }
}


int is_keyword(struct Token const *t) {
    unsigned char const *k;
    unsigned char const *r = t->text + 1;
    size_t               L = t->length - 1;
    struct Keyword_Tail *p;
    size_t               i;

    if (t->length == 0) return 0;
    if (keyword_set.case_sensitive) {
	for (p = keyword_set.words[r[-1]]; p != 0; p = p->next) {
	    if (p->size == L) {
		k = p->tail;
		for (i = 0; i < L; i++) {
		    if (r[i] != k[i]) break;
		}
		if (k[i] == '\0') return p->kind;
	    }
	}
    } else {
	for (p = keyword_set.words[dncase[r[-1]]]; p != 0; p = p->next) {
	    if (p->size == L) {
		k = p->tail;
		for (i = 0; i < L; i++) {
		    if (dncase[r[i]] != k[i]) break;
		}
		if (k[i] == '\0') return p->kind;
	    }
	}
    }	
    return 0;
}



static char *m4_keywords[] = {
    "assert",
    "define",
    "ifdef",
    "ifelse",
    "include",
    "popdef",
    "pushdef",
    "sinclude",
    "undefine",
    (char*)0
};


void init_ctype(int M4_used, struct Language const *l) {
    int c;
    unsigned char const *p;
    char **k;

    for (c = 0; c < 256; c++) {
	ctype[c]  = 0;
	nocase[c] = c;
	dncase[c] = c;
	upcase[c] = c;
    }	

    for (c = '0'; c <= '9'; c++) {
	ctype[c] |= DIGIT;
    }

    for (c = 'A'; c <= 'Z'; c++) {
	ctype[c] |= UPPER;
	upcase[c+32] = c;
    }
    for (c = 0xc0; c < 0xd7; c++) {
	ctype[c] |= UPPER;
	upcase[c+32] = c;
    }
    for (c = 0xd8; c < 0xdf; c++) {
	ctype[c] |= UPPER;
	upcase[c+32] = c;
    }

    for (c = 'a'; c <= 'z'; c++) {
	ctype[c] |= LOWER;
	dncase[c-32] = c;
    }
    ctype[0xdf] |= LOWER;	/* upcase[sharp s] == sharp s */
    for (c = 0xe0; c < 0xf7; c++) {
	ctype[c] |= LOWER;
	dncase[c-32] = c;
    }
    for (c = 0xf8; c < 0xff; c++) {
	ctype[c] |= LOWER;
	dncase[c-32] = c;
    }
    ctype[0xff] |= LOWER;	/* upcase[y umlaut] == y umlaut */

    if (M4_used) {
	ctype['_'] |= UNDERS;
	ctype['$'] |= UNDERS;
	add_kinded_keyword("dnl", 3, -1);
	for (k = m4_keywords; *k; k++)
	    add_m4_keyword(*k);
    }

    escaper = '\0';
    opchar  = '\0';

    if (l != 0) {
	for (p = (unsigned char const *)l->lbracks; *p; p++)
	    ctype[*p] |= LBRACK; 
	for (p = (unsigned char const *)l->rbracks; *p; p++)
	    ctype[*p] |= RBRACK;
	for (p = (unsigned char const *)l->unders; *p; p++)
	    ctype[*p] |= UNDERS;
	for (p = (unsigned char const *)l->specials; *p; p++)
	    ctype[*p] |= SPECLS;
	escaper = l->escaper;
	opchar  = l->opchar;

	for (k = l->keywords; *k; k++)
	    add_keyword(*k);
	keyword_set.case_sensitive = l->case_sensitive;
    }
}


static struct {
    char c1, c2; struct Language const *l;
} opt_table[] = {
    { 'a', 'd',  &ada     },
    { 'a', '\0', &ada     },
    { 'a', 'w',  &awk     },
    { 'c', '+',  &cpp     },
    { 'c', 'p',  &cpp     },
    { 'c', 'x',  &cpp     },
    { 'c', '8',  &c89     },
    { 'c', '\0', &c89     },
    { 'd', '\0', &d04     },
    { 'd', 't',  &sgml    },
    { 'e', 'i',  &eiffel  },
    { 'e', 'n',  &sgml    },
    { 'e', 'r',  &erlang  },
    { 'e', 's',  &eiffel  },
    { 'e', '\0', &eiffel  },
    { 'h', 'x',  &cpp     }, /* hxx */
    { 'h', 't',  &sgml    }, /* htm[l] */
    { 'h', 's',  &haskell }, 
    { 'h', 'a',  &haskell },
    { 'h', 'p',  &cpp     }, /* hpp */
    { 'h', '\0', &c89     },
    { 'j', 'a',  &java    },
    { 'j', '\0', &java    },
    { 'm', 'e',  &mercury },
    { 'm', 'l',  &sml     },
    { 'm', '\0', &mercury },
    { 'n', 'q',  &nqc     },
    { 'o', 'p',  &opl     },
    { 'p', 'a',  &pascal  },
    { 'p', 'r',  &prolog  },
    { 'p', 'l',  &prolog  },
    { 'p', '\0', &pascal  },
    { 'r', 'a',  &raw     },
    { 'r', 'p',  &rpcl    },
    { 'r', '\0', &raw     },
    { 's', 'g',  &sgml    },
    { 's', 'm',  &sml     },
    { 's', 'q',  &sql     },
    { 't', 'e',  &raw     },
    { 't', 'x',  &raw     },
    { 't', '\0', &raw     },
    { 'x', 'm',  &sgml    },
    { 's', 't',  &st80    },
    { 'c', 's',  &st80    },
    { 's', 'c',  &scm     },
    { 's', '\0', &scm     },
    { 'R', '\0', &R       },
    { 'S', '\0', &R       },
#if 0    
    { 'c', 'l',  &clean   },
    { 'd', 'c',  &clean   },
    { 'i', 'c',  &clean   },
    { 'i', 'd',  &idl     },
    { 'i', '\0', &idl     },
    { 'l', 'i',  &lisp    },
    { 'l', 's',  &lisp    },
    { 'l', '\0', &lisp    },
    { 'p', 'e',  &perl    },
    { 'x', '\0', &rpcgen  },
#endif
    { '\0','\0', 0        }
};


struct Language const *opt_lang(char const *opt) {
    unsigned char const c1 = dncase[(unsigned char)opt[0]];
    unsigned char const c2 = dncase[(unsigned char)opt[1]];
    int i;

    for (i = 0; opt_table[i].c1 != '\0'; i++)
	if (opt_table[i].c1 == c1 && opt_table[i].c2 == c2)
	    return opt_table[i].l;

    fprintf(stderr, "m2h: unknown language -l%s\n", opt);
    fprintf(stderr, "known languages are\n");
    for (i = 0; opt_table[i].c1 != '\0'; i++)
	fprintf(stderr, "    -l%c%c  %s\n",
		opt_table[i].c1,
		(opt_table[i].c2 == '\0' ? ' ' : opt_table[i].c2),
		opt_table[i].l->name);
    exit(EXIT_FAILURE);
    return &raw;
}


static struct Language const *ext_table[] = {
    &ada, &awk, &c89, &cpp, &d04, &eiffel, &erlang, &haskell, &java,
    &mercury, &nqc, &opl, &pascal, &prolog, &R, &raw, &rpcl, &scm,
    &sgml, &sml, &sql, &st80,
#if 0
    &clean, &idl, &lisp, &perl,
#endif
    0
};


struct Language const *ext_lang(char const *title) {
    int i;
    char const *e;

    for (e = title; *e != '\0'; e++) {}
    for (;; e--) {
	if (e == title) return &raw;	/* no extension */
	i = e[-1];
	if (i == '/')   return &raw;	/* Unix  directory separator */
	if (i == '\\')  return &raw;	/* MSDOS directory separator */
        if (i == '.')   break;
    }
    if (*e == '\0')     return &raw;	/* empty extension */

    for (i = 0; ext_table[i] != 0; i++) {
	char const *s = ext_table[i]->suffixes;

	while (*s == '.') {
	    unsigned char const *f, *t;
	    s++;
	    f = (unsigned char const *)e;
	    t = (unsigned char const *)s;
	    do {
		if (*f == '\0' && (*t == '\0' || *t == '.'))
		    return ext_table[i];
	    } while (dncase[*f++] == dncase[*t++]);
	    while (*s != '.' && *s != '\0') s++;
	}
    }
    fprintf(stderr, "m2h: unknown extension .%s of %s\n", e, title);
    fprintf(stderr, "known extensions are\n");
    for (i = 0; ext_table[i] != '\0'; i++)
	fprintf(stderr, "    %s   (%s)\n",
		ext_table[i]->suffixes,
		ext_table[i]->name);
    exit(EXIT_FAILURE);
    return &raw;
}


#if TEST

int main(int argc, char **argv) {
    int i;
    struct Language const *l;

    init_ctype(0, 0);
    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    l = opt_lang(argv[i] + (argv[i][1] == 'l' ? 2 : 1));
	} else {
	    l = ext_lang(argv[i]);
	}
	printf("%2d. %s => %s\n", i, argv[i], l->name);
    }
    return 0;
}

#endif

