/*  File   : prolog.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for Prolog and Erlang.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 prolog.c	1.14";
#endif/*lint*/

#include "many2h.h"
#include "language.h"


/*  There are four string-like things we have to deal with:
    "true strings"
    'quoted atoms'
    0'c Prolog character literals
    $c Erlang character literals
    In every other language there is an assumption that an identifier
    cannot be split across line boundaries.  This is false in Prolog
    and Erlang.  I've decided to adopt a compromise:  if a quoted atom
    is complete on one line, it is treated as an identifier, otherwise
    all of it will be treated as a string.  This means that it is now
    possible to have <o>' <i>foobar <q>' where we would normally expect
    <p>foobar.

    For my convenience, both Prolog character literals and Erlang
    character literals are recognised.  This is correct for neither
    language, but should be harmless.

    These languages are case sensitive, with upper case indicating
    variables and lower case indicating constants, constructors, and
    predicate/function names.  I've decided that variable names will
    be treated as Attributes and the others as Identifiers.
*/
static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = is_lower(c) ? Identifier : Attribute;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {

		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    token_row[t].type   = Number;
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (column - newcol == 1 && line_buffer[newcol] == '0') {
	        if (c == 'o') {
	            /* ISO Prolog hates Edinburgh and loves C */
	            column++;
	            while (is_a_digit(line_buffer[column]))
			column++;
		    c = ' ';
		} else
		if (c == 'x') {
		    /* ISO Prolog hates Edinburgh and loves C */
		    column++;
		    while (is_a_letter_or_digit(line_buffer[column]))
		    	column++;
		    c = ' ';
		} else
		if (c == '\'') {
		    c = line_buffer[++column];
		    if (c == '\\') {
			c = line_buffer[++column];
			if (is_a_digit(c)) {
			    do column++; while (
				is_a_digit(line_buffer[column]));
			} else
			if (dncase[c] == 'x') {
			    do column++; while (
				is_a_letter_or_digit(line_buffer[column]));
			} else {
			    column++;
			}
		    } else {
		        column++;
		    }
		    c = ' ';
		    token_row[t].type = Character_Name;
		}
	    }
	    if (c == '\'') {
                do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    } else
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
		if (dncase[line_buffer[column]] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit_or_underscore(line_buffer[column]))
			column++;
		}
	    }
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '%') {
	    token_row[t].type = Comment_Opener;
	    do column++; while (line_buffer[column] == c);
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '/' && line_buffer[column+1] == '*') {
	    static unsigned char const ender[3] = {'*', '/', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 2;
	    column += 2;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 2;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '"' || c == '\'') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        if (c == '\'') token_row[t].type = Identifier;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (c == '$') {
	    column++;
	    c = column < line_length ? line_buffer[column++] : ' ';
	    if (c == '\\') {
		c = column < line_length ? line_buffer[column++] : ' ';
		if (is_a_digit(c)) {
		    while (is_a_digit(line_buffer[column])) column++;
		} else
		if (dncase[c] == 'x') {
		    while (is_a_letter_or_digit(line_buffer[column])) column++;
		}
	    }
	    token_row[t].type = Character_Name;
	    token_row[t].length = column - token_row[t].start;
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == ',' || c == ';' || c == '|') {
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    if (c == '.' && line_buffer[column+1] <= ' ') {
		token_row[t].type   = Dot;
		token_row[t].length = 1;
		column++;
	    } else {
		do {
		    column++;
		} while (is_a_special_character(line_buffer[column]));
		token_row[t].type   = Special;
		token_row[t].length = column - token_row[t].start;
	    }
	    state = Other_Token;
	} else {
	    /* @ $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *prolog_keywords[] = {
    "bagof",
    "div",
    "dynamic",
    "fail",
    "false",
    "findall",
    "foreign",
    "is",
    "meta_predicate",
    "mod",
    "mode",
    "module",
    "multi_file",
    "nospy",
    "not",
    "on_exception",
    "op",
    "otherwise",
    "pred",
    "public",
    "raise_exception",
    "setof",
    "spy",
    "true",
    "type",
    "use_module",
    "volatile",
    (char *)0
};

struct Language prolog = {
    "Prolog",
    ".pl.nl",
    "([{",
    ")]}",
    "_",
    "~!#^&@*-+=|\\<>?:/.",
    '\\',
    '\0',    /* user-defined operators are done another way */
    Is_Case_Sensitive,
    prolog_keywords,
    tokenise
};


static char *mercury_keywords[] = {
    "bagof",
    "div",
    "dynamic",
    "else",
    "end_module",
    "fail",
    "false",
    "findall",
    "foreign",
    "func",
    "if",
    "implementation",
    "import_module",
    "instance",
    "interface",
    "is",
    "lambda",
    "meta_predicate",
    "mod",
    "mode",
    "module",
    "multi_file",
    "not",
    "on_exception",
    "op",
    "otherwise",
    "pragma",
    "pred",
    "raise_exception",
    "setof",
    "then",
    "true",
    "type",
    "typeclass",
    "volatile",
    "where",
    (char *)0
};

struct Language mercury = {
    "Mercury",
    ".me.m",
    "([{",
    ")]}",
    "_",
    "~!#^&@*-+=|\\<>?:/.",
    '\\',
    '\0',    /* user-defined operators are done another way */
    Is_Case_Sensitive,
    mercury_keywords,
    tokenise
};


static char *erlang_keywords[] = {
    "after",
    "andalso",
    "band",
    "begin",
    "bnot",
    "bor",
    "bsl",
    "bsr",
    "bxor",
    "case",
    "catch",
    "cond",
    "div",
    "end",
    "export",
    "fun",
    "if",
    "import",
    "module",
    "of",
    "orelse",
    "receive",
    "rem",
    "when",
    (char *)0
};

struct Language erlang = {
    "Erlang",
    ".erl.hrl.yrl",
    "([{",
    ")]}",
    "_",
    "~!#^&@*-+=|\\<>?:/.",
    '\\',
    '\0',    /* no user-defined operators */
    Is_Case_Sensitive,
    erlang_keywords,
    tokenise
};


