/*  File   : st80.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for Smalltalk-80
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 st80.c	1.5";
#endif/*lint*/

#include "many2h.h"
#include "language.h"

/*  I want contextual handling of tokens.
    If a line begins with a letter, I'll assume it is a class or method
    declaration, so I want words to be reported as Keywords.

    After a sharp, I want the tokens forming the literal to be reported
    as Attributes.

    To do this, we need two variables:
    context < 0 => in method header.  Reset by any non-word.
    context > 0 => in a literal.  Reset at end of literal.
    context = 0 => normal situation.

    depth = bracket nesting count, but only active when context > 0.
*/
static int context = 0;
static int depth   = 0;
/*  Smalltalk tokens are
    - punctuation
      * Left_Bracket	( [ {
      * Right_Bracket   } ] )
      * Special         . : ; ^ |
    - comments "..." (can span lines, no doubling)
      Comment_Opener "
      Comment_Element inner stuff (!! mapped to !)
      Comment_Closer "
    - strings '...' (can span lines, doubling '')
      String_Opener '
      String_Element inner stuff (!! mapped to !, '' taken as content)
      String_Closer '
    - characters $<char>
      Character_Name
    - numbers <base>r[-]<xdigit>+, <digit>+, <digit>*.<digit>+,
      <digit>+e<sign>?<digit>+ (case significant),
      <digit>*.<digit>+e<sign>?<digit>+ (case significant)
      Number
    - words, of three kinds:
      * (<letter><letdig>*:)+		(all or part of keyword selector)
      * <letter><letdig>*		(unary, or variable)
      * - | <symbol><symbol>?		(binary)
        where <symbol> is +/\*~<>=@%&?, or !!
      Identifier, Keyword, or Attribute when context =, <, > 0.
    - one of three special marks:
      * _ is to be converted to :=
	Special
      * !! is to be converted to plain ! (inside string or comment or word)
        ! on its own is to be bold
	Special (! on its own only)
      * # introduces a literal.
	Tick.
*/
static void tokenise(void) {
    int column, newcol;
    int t;
    unsigned char c;
    int shift = 0;

    if (is_a_letter(line_buffer[0])
     && state != String_Open && state != Comment_Open
    ) {   
        context = -1;
    } else
    if (context < 0) {
	context = 0;
    }

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, terminator[0]);
	token_row[0].type   = context > 0 ? Attribute : String_Element;
	token_row[0].start  = column + shift;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type   = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = context > 0 ? Attribute : String_Closer;
	token_row[1].start  = newcol + shift;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = Term_Ender;
	t = 2;
	if (context != 0 && depth == 0) context = 0;
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column + shift;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol + shift;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column + shift;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    while (line_buffer[column] == '_' &&
	           is_a_letter_or_digit(line_buffer[column+1])
	    ) {
		do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    }
	    token_row[t].length = column + shift - token_row[t].start;
	    token_row[t].type = context < 0 ? Keyword
                              : context > 0 ? Attribute : Identifier;
	    if (context > 0 && depth == 0) context = 0;
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    token_row[t].type   = context > 0 ? Attribute : Number;
	    newcol = column;
	    do column++;
		while (is_a_digit(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == 'r') {
		if (line_buffer[column+1] == '-') column++;
		do column++; while (
		    is_a_letter_or_digit(line_buffer[column]));
	    } else
	    if (c == '.') {
		do column++;
		    while (is_a_digit(line_buffer[column]));
		if (line_buffer[column] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit(line_buffer[column]))
			column++;
		}
	    }
	    token_row[t].length = column + shift - token_row[t].start;
	    if (context != 0 && depth == 0) context = 0;
	    state = Term_Ender;
	} else
	if (c == '"') {
	    static unsigned char const ender[2] = {'"', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 1;
	    column++;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 1;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column + shift;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol + shift;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '\'') {
	    static unsigned char ender[2] = {'\'', '\0'};

	    token_row[t].type   = context > 0 ? Attribute : String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = context > 0 ? Attribute : String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column + shift;
	    newcol = terminator_column(column, c);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        token_row[t].type   = context > 0 ? Attribute : String_Element;
	        t++;
	        token_row[t].type   = context > 0 ? Attribute : String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol + shift;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
	        if (context != 0 && depth == 0) context = 0;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length + shift;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (c == '$') {
	    column++;
	    c = column < line_length ? line_buffer[column++] : ' ';
	    token_row[t].type = context > 0 ? Attribute : Character_Name;
	    token_row[t].length = column + shift - token_row[t].start;
	    if (context != 0 && depth == 0) context = 0;
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = context > 0 ? Attribute : Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    if (context > 0) depth++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = context > 0 ? Attribute : Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    if (context > 0) {
		if (depth > 0) depth--;
		if (depth == 0) context = 0;
	    }
	    state = Term_Ender;
	} else
	if (c == '.' || c == ':' || c == ';'
         || c == '^' || c == '|'
        ) {
	    token_row[t].type    = Special;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '#') {
	    token_row[t].type   = Tick;
	    token_row[t].length = 1;
	    column++;
	    context = 1;		/* WAS -1 */
	    depth   = 0;
	    state = Other_Token;
	} else
	if (c == '_') {
	    token_row[t].text   = (unsigned char const *)":= ";
	    token_row[t].length = 2;
	    token_row[t].type   = Special;
	    /* We need to lie about the starting point, in order to get
	       the output right.
	       for a space preceding and a space following, and lie a
	       little more.
	    */
	    shift++;	    
	    if (column > 0 && line_buffer[column-1] > ' ')
		token_row[t].start++, shift++;
	    if (line_buffer[column+1] > ' ') shift++;
	    column++;
	} else
	if (is_a_special_character(c)
	 || (c == '!' && line_buffer[column+1] == '!')
	) {
	    if (c == '!') shift--, column += 2;
	    else column++;
	    c = line_buffer[column];
	    if (is_a_special_character(c)
	     || (c == '!' && line_buffer[column+1] == '!')
	    ) {
		column += c == '!' ? 2 : 1;
		token_row[t].length = 2;
	    } else {
		token_row[t].length = 1;
	    }
	    token_row[t].type   = context > 0 ? Attribute : Special;
	    if (context > 0 && depth == 0) context = 0;
	    state = Other_Token;
	} else
	if (c == '!') {
	    token_row[t].type = Keyword;
	    token_row[t].length = 1;
	    column++;
	    context = 0; depth = 0;
	    state = Other_Token;
	} else {
	    /* @ $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column + shift;
    token_row[t].length = 0;
}


static char *st80_keywords[] = {
    "and:",
    "ifFalse:",
    "ifNil:",
    "ifNotNil:",
    "ifTrue:",
    "or:",
    "whileFalse",
    "whileFalse:",
    "whileTrue",
    "whileTrue:",
    (char *)0
};

struct Language st80 = {
    "Smalltalk",
    ".cs.st",
    "([{",
    ")]}",
    ":",
    "+-/\\*~<>=@%&?,",
    '\'',
    '\0',
    Is_Case_Sensitive,
    st80_keywords,
    tokenise
};


