/*  File   : formatter.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Stuff shared by all formatters.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 formatter.c	1.14";
#endif/*lint*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "formatter.h"


/*  Each formatter must provide

    .start_style
	a table mapping styles to markup for starting
    .end_style
	a table mapping styles to markup for ending
    .char_map
	a table mapping each character to a markup-immune string
	e.g. '&' -> "&amp" for HTML or '$' -> "\\$" for LaTeX.
    .begin_output(author, date, title1, title2)
	a function to generate the beginning of a document;
	author is a string saying who ran many2h;
	date is a string giving the time many2h ran in ISO 8601 format;
	title1 is the first part of the title;
	title2 is the second part of the title.
	If a string argument is null or empty, that information is not
	present and should not be used.
    .end_output()
	a function to wrap up the end of a document.
    .begin_section(title)
	a function to begin a new section, corresponding to a new source
	file.  The title is the name of the file.  It will be null or
	empty when the input comes from stdin.
    .end_section()
	a function to end a section.
    .tab_from_to(from_column, to_column)
	a function to emit spaces (for fixed width output) or to
	move to to_column (for variable width output).
    .finish_line()
	a function to finish the current output line

    This file provides support operations

    put_spaces(n)
	write n spaces to stdout
	
    put_plain(s)
	write string s to stdout, like fputs(s, stdout) but maybe faster.

    put_mapped(s, &formatter)
	write string s to stdout with characters mapped for formatter.

    put_styled(s, &formatter, style)
	write string s to stdout with appropriate markup to get the
	requested style (plain, bold, &c) and with characters mapped
	for formatter.

    write_chunk(s, n, &formatter, style)
	the same as put_styled() except that the string has n characters;
	it is not NUL-terminated.  I intend that this function should
	go away.

    To use a formatter, you do

	f->begin_output(author, date, title1, title2);
	for each source file (title) {
	    open the file
	    f->begin_section(title);
	    for each line of the file {
		tokenise the line
		for each token {
		    f->tab_from_to(last column, new column);
		    write_chunk(ptr, len, f, style);
		    f->end_token(the type of token)
		}
		f->finish_line();
	    }
	    f->end_section();
	}
	f->end_output();

    By avoiding printf() and fputs(), I have managed to save about 1/3rd
    of the time that the earlier monolithic version of this program took.
    In order to accomplish this, the functions put_spaces(), put_plain(),
    and put_mapped() were written to exploit my knowledge of the internals
    of the traditional UNIX stdio implementation.  However, such code is
    not portable.  My mistake was not bypassing stdio *entirely*.  If any
    of the code in this file is used, *all* output to stdout goes through
    here, so we can just use write() directly.  The one thing we lose by
    doing so is automatic flush at the end, hence flush_output is now
    public and called by core.c.
*/

#define OUTPUT_BUFFER_SIZE 8192
static unsigned char  buffer[OUTPUT_BUFFER_SIZE];
static unsigned char *ptr = &buffer[0];
static size_t         cnt = OUTPUT_BUFFER_SIZE;


void flush_output(void) {
    unsigned char const *p = buffer;
    size_t               w = ptr - buffer;
    ssize_t              r;

    while (w != 0) {
	r = write(STDOUT_FILENO, p, w);
	if (r < 0) {
	    perror("m2h: error writing stdout");
	    exit(EXIT_FAILURE);
	}
	if (r == 0) {
	    abort();		/* This should not happen */
	}
	w -= (size_t)r;
    }
    ptr = buffer;
    cnt = OUTPUT_BUFFER_SIZE;
}


void put_spaces(int n) {
    if (n <= 0) return;
    for (;;) {
	register unsigned char *d = ptr;
	register size_t         c = cnt;

	while (c-- != 0) {
	    *d++ = ' ';
	    if (0 == --n) {
		cnt = c;
		ptr = d;
		return;
	    }
        }
        ptr = d;
        flush_output();
    }
}


void put_plain(register unsigned char const *s) {
    register unsigned char x;

    if ((x = *s++) == '\0') return;
    for (;;) {
	register unsigned char *d = ptr;
	register size_t         c = cnt;

	while (c-- != 0) {
	    *d++ = x;
	    if ((x = *s++) == '\0') {
	        ptr = d;
	        cnt = c;
		return;
	    }
        }
	ptr = d;
	flush_output();
    }
}


/*  My plan to switch from write_chunk() to put_styled() has never come
    to fruition.  Maybe it shouldn't.  If not, then this function should
    be rewritten to take arguments t, n, f (pointer, length, formatter).
    However, it is currently called with its present interface.
*/
void put_mapped(unsigned char const *t, struct Formatter const *const f) {
    register unsigned char x;
    register unsigned char const *s;
    char const *const * m = f->char_map;

    do {
	if ((x = *t++) == '\0') return;
	s = (unsigned char const *)m[x];
    } while (s == 0 || (x = *s++) == '\0');

    for (;;) {
	register unsigned char *d = ptr;
	register size_t         c = cnt;

	while (c-- != 0) {
            *d++ = x;
	    if ((x = *s++) == '\0') {
		do {
		    if ((x = *t++) == '\0') {
			ptr = d;
			cnt = c;
			return;
		    }
		    s = (unsigned char const *)m[x];
		} while (s == 0 || (x = *s++) == '\0');
	    }
        }
        ptr = d;
        flush_output();
    }
}

#if 0

void put_spaces(int n) {
    FILE *p = stdout;

    while (n-- > 0) put_char(' ', p);
}


void put_plain(register unsigned char const *s) {
    FILE *p = stdout;
    register unsigned char x; /* character to write */

    while ((x = *s++) != '\0') put_char(x, p);
}


void put_mapped(unsigned char const *t, struct Formatter const *const f) {
    FILE                         *p = stdout;
    register unsigned char        x; /* character to write */
    register unsigned char const *s;
    char const *const *           m = f->char_map;

    while ((x = *t++) != '\0') {
	s = (unsigned char const *)m[x];
        if (s != 0) {
	    while ((x = *s++) != '\0') put_char(x, p);
    }
}

#endif


void put_styled(
    unsigned char const *s,
    struct Formatter const *const f,
    enum Text_Style style
) {
    put_string(f->start_style[style]);
    put_mapped(s, f);
    put_string(f->end_style[style]);
}


void write_chunk(
    unsigned char const *s,
    size_t n,
    struct Formatter const *const f,
    enum Text_Style style
) {
    size_t i;

    put_string(f->start_style[style]);
    for (i = 0; i!= n; i++) put_string(f->char_map[s[i]]);
    put_string(f->end_style[style]);
}

		  /*ARGSUSED*/
void no_end_token(enum Token_Type type) {
    (void)type;
}


struct Formatter const *opt_fmt(char const *opt) {
    switch (opt[0]) {
    case 'p': case 'P':
	return &plain;
    case 'a': case 'A':
	return &ansi;
    case 'c': case 'C':
	return &colour;
    case 'h': case 'H':
	return &html;
    case 'f': case 'F':
        return &chtml;
    case 'l': case 'L':
	return &latex;
    case 'r': case 'R':
	return &roff;
    case 'o': case 'O': case '1':
        return &one_per_line;
    default:
	fprintf(stderr, "many2h: unknown formatter -f%s\n", opt);
	fprintf(stderr, "known formatters are\n");
	fprintf(stderr, "    -fp  Plain text (default)\n");
	fprintf(stderr, "    -fa  ANSI terminal bold/underline\n");
	fprintf(stderr, "    -fc  ansi terminal Colouring\n");
	fprintf(stderr, "    -fh  HTML with black-and-white highlights\n");
	fprintf(stderr, "    -ff  HTML with coloured highlights\n");
	fprintf(stderr, "    -fl  LaTeX\n");
	fprintf(stderr, "    -fr  Troff or nroff\n");
	fprintf(stderr, "    -fo  One token per line, plain text\n");
	exit(EXIT_FAILURE);
	return &plain;
    }
}


#if TEST

#include <string.h>

static void convert_using(struct Formatter const *const f) {
    int col;
    int c;

    f->begin_output("Tester", "1999-06-12T17:01:35+00:00",
	"Converted plain text", "");
    f->begin_section("");
    col = 0;
    while ((c = getchar()) >= 0) {
	if (c == '\n') {
	    f->finish_line();
	    col = 0;
	} else
	if (c == '\t') {
	    do put_string(f->char_map[' ']); while (++col % 8 != 0);
        } else {
	    put_string(f->char_map[c]);
            col++;
	}
    }
    f->end_section();
    f->end_output();
}


int main(int argc, char **argv) {
    struct Formatter const *f;
    char *x;

    x = argc > 1 ? argv[1] : argv[0];
    while (*x == '-' || *x == '/' || *x == '\\') x++;
    if (strchr(x, '/') != 0) x = 1+strchr(x, '/');
    f = opt_fmt(x);
    convert_using(f);
    return 0;
}

#endif
